#include "MtQueue.h"
#include "db_dbs_rx.h"
#include <usrp_standard.h>
#include <fstream>
#include <iostream>
//#include <pthread.h>

using namespace std;

void* Producer(void* Args);
void* Consumer(void* Args);

static const unsigned n = 4096*2;
typedef TCMtQueue<short[n],4096> TDQueue;
//-----------------------------------------------------------------------------
// TSProducerArg
//
// Provides the queue and the USRP to the producer thread.
//-----------------------------------------------------------------------------
struct TSProducerArg
{
  TDQueue* pQueue;
  usrp_standard_rx* pRx;
};
//-----------------------------------------------------------------------------
// TSConsumerArg
//
// Provides the queue and the output stream to the consumer thread.
//-----------------------------------------------------------------------------

struct TSConsumerArg
{
  TDQueue* pQueue;
  ostream* pStream;
};
//-----------------------------------------------------------------------------
// Producer
//
// This thread reads data from the USRP as fast as possible and places the
// data onto the queue.  If the queue fills up, data will not be read 
// from the USRP as this code blocks in TCMtQueue::Add.  The USRP may become
// upset about this and overrun
//-----------------------------------------------------------------------------
void* Producer(void* q)
{
  TSProducerArg* pArg = reinterpret_cast<TSProducerArg*>(q);
  TDQueue *pQueue = pArg->pQueue;
  usrp_standard_rx* rx = pArg->pRx;

  rx->start();

  while(true)
  {
    bool Overrun;
    short Buffer[n];

    int NumBytes = rx->read(
      (char*)Buffer,
      n*sizeof(short),
      &Overrun);

    if(NumBytes == -1)
    {
      cerr << "Problem" << endl;
    }
    else
    {
      if(Overrun)
      { 
        cerr << "overrun" << endl;
      }

      // Throw away the first 10000 samples because the first thousand or
      // so samples from the USRP are bogus while it warms up.
      static bool SkipFirstSamples = true;

      if(SkipFirstSamples)
      {
        static unsigned NumSkipped = 0;

        NumSkipped += n;

        if(NumSkipped >= 10000)
        {
          SkipFirstSamples = false;
        }
      }
      else
      {
        short* Temp = pQueue->Add1();
        memcpy(Temp, Buffer, sizeof(Buffer));
        pQueue->Add2();
      }
    }
  }

  rx->stop();

  return 0;
}
//-----------------------------------------------------------------------------
// Consumer
//
// Thread that writes data from the queue to the stream as fast as possible.
// If this thread cannot write data fast enough, the queue will fill up
// and the producer thread will block.  If this thread writes too quickly,
// it will block on TCMtQueue::Delete and await more data from the producer.
//-----------------------------------------------------------------------------
void* Consumer(void* q)
{
  TSConsumerArg* pArg = reinterpret_cast<TSConsumerArg*>(q);
  TDQueue *pQueue = pArg->pQueue;
  ostream* pStream = pArg->pStream;

  while(true)
  {
     short* Buffer = pQueue->Delete1();
     pStream->write((char*)Buffer, n*sizeof(short));
     pQueue->Delete2();
  }

  return 0;
}
//-----------------------------------------------------------------------------
// main
//
// Parse the command line.  Prep the USRP with frequencies/decimiation rates/
// etc.  Launch the producer/consumer threads and start running.
//-----------------------------------------------------------------------------
int main(int argc, char** argv)
{
  enum TEBoard      { eDbsrx, eBasicRx } Board      = eBasicRx;
  enum TESampleSize { e16Bit, e8Bit    } SampleSize = e16Bit;

  int NumChannels = 1;

  for(int i = 1; i < argc; ++i)
  {
    std::string String = argv[i];

    for(unsigned j = 0; j < String.size(); ++j)
    {
      char c = String[j];

      switch(c)
      {
      case 'd':
        Board = eDbsrx;
        break;
      case 'b':
        Board = eBasicRx;
        break;
      case '1':
        NumChannels = 1;
        break;
      case '2':
        NumChannels = 2;
        break;
      case '8':
        SampleSize = e8Bit;
        break;
      case '6':
        SampleSize = e16Bit;
        break;
      };
    }
  }

  static const int UsrpNumber = 0;

  const unsigned Format = usrp_standard_rx::make_format(
    (SampleSize == e8Bit ? 8 : 16), // width across usb
    (SampleSize == e8Bit ? 8 : 0),  // shift
    true,                           // want Q
    false                           // bypass halfband
    );

  usrp_standard_rx* rx = usrp_standard_rx::make(UsrpNumber, 16);

  if(Board == eBasicRx)
  {
    cerr << "BasicRX\n";
    cerr << NumChannels << " channel"
         << (NumChannels == 1 ? "" : "s")
         << "\n";

    static const int Mux = 0x33221100;

    static const unsigned DecimRate = 16;
    rx->set_decim_rate(DecimRate);

    static const double DdcFreq = 10000;

    rx->set_nchannels(NumChannels);

    for(int i = 0; i < NumChannels; ++i)
    {
      rx->set_rx_freq(i, DdcFreq);
    }

    rx->set_mux(Mux);
    rx->set_pga(0, 0);
    rx->set_pga(1, 0);
  }
  else if(Board == eDbsrx)
  {
    cerr << "DBSRX\n";

    db_dbs_rx dbsrx(rx, UsrpNumber);
    static const double Freq = 1.57542e9;

    //  Add 600kHz to remove birdie
    static const double BirdieOffset = 600e3;
    double TunedFreq = dbsrx.tune(Freq + BirdieOffset);

    if(!TunedFreq)
    {
      cerr << "Could not tune the DBSRX.  Is it plugged in?" << endl;
    }

    dbsrx.gain(50);

    double DdcFreq = TunedFreq-Freq;
    static const int Mux = 0x00000010;
    static const unsigned DecimRate = 16;
    rx->set_decim_rate(DecimRate);
    rx->set_rx_freq(0, DdcFreq);
    rx->set_mux(Mux);
  }

  rx->set_format(Format);

  TDQueue* pQueue = new TDQueue;
  pthread_t pro, con;

  TSProducerArg ProdArg = {pQueue, rx};
  TSConsumerArg ConsArg = {pQueue, &(std::cout)};

  pthread_create (&pro, NULL, Producer, &ProdArg);
  pthread_create (&con, NULL, Consumer, &ConsArg);
  pthread_join (pro, NULL);
  pthread_join (con, NULL);

  delete pQueue;

  return 0;
}
